#include <os.h>

#define K 1024
#define M (K * K)

#define VIDEO_BASE           0xB8000

unsigned long mem_end;          // Size of memory
char *heap;                     // Heap pointer point to next free page in memory
pte *pdir;                      // Page directory
struct syspage *syspage;        // System page with tss and gdt
pte *syspagetable;              // System page table
char *initstk;                  // Initial stack for kernel
unsigned long krnlentry;        // Virtual address of kernel entry point
void *physgdt;			// Physical address of GDT
void *physidt;			// Physical address of IDT
unsigned short tssval;          // TSS selector value
unsigned short ldtnull;         // LDT null selector
struct selector gdtsel;         // GDT selector
struct selector idtsel;         // GDT selector

void panic(char *msg)
{
  kprintf("Panic: %s\n", msg);
  while (1);
}

char *alloc_pages(int numpages)
{
  char *p;

  // check for out of memory
  if ((unsigned long) heap + numpages * PAGESIZE >= mem_end) panic("out of memory");
  p = heap;

  // zero allocated pages
  memset(p, 0, numpages * PAGESIZE);

  // update heap pointer
  heap += numpages * PAGESIZE;

  return p;
}

unsigned long memsize()
{
  volatile unsigned long *mem;
  unsigned long addr;
  unsigned long value;
  unsigned long cr0save;
  unsigned long cr0new;

  // start at 1MB
  addr = 1 * M;

  // save a copy of CR0
  __asm { mov eax, cr0 };
  __asm { mov [cr0save], eax };

  // invalidate the cache (write-back and invalidate the cache)
  __asm { wbinvd };

  // plug cr0 with just PE/CD/NW (cache disable(486+), no-writeback(486+), 32bit mode(386+))
  cr0new = cr0save | 0x00000001 | 0x40000000 | 0x20000000;
  __asm { mov eax, [cr0new] };
  __asm { mov cr0, eax };

  // probe for each megabyte
  while (addr < 0xFFF0000000000)
  {
    //kprintf("%04d MB RAM\r", addr / M);

    addr += 1 * M;
    mem= (unsigned long *) addr;

    value = *mem;
    *mem = 0x55AA55AA;

    if (*mem != 0x55AA55AA) break;

    *mem = 0xAA55AA55;
    if(*mem != 0xAA55AA55) break;

    *mem = value;
  }

  // restore 
  __asm { mov eax, [cr0save] };
  __asm { mov cr0, eax };

  //kprintf("%04d MB RAM\n", addr / M);
  return addr;
}
	
void check_modules(char *base)
{
  struct dos_header *doshdr;
  struct image_header *imghdr;

  // check that kernel does not overlap data section in osldr
  doshdr = (struct dos_header *) base;
  imghdr = (struct image_header *) (base + doshdr->e_lfanew);

  if ((unsigned long) doshdr->e_cp * PAGESIZE != imghdr->optional.size_of_image)
  {
    panic("kernel overlaps os loader data section");
  }
}

void copy_kernel(char *base)
{
  struct dos_header *doshdr;
  struct image_header *imghdr;
  int kernelsize;
  int kernelpages;
  char *kerneladdr;
  pte *pt;
  int i;

  // locate os loader image header
  doshdr = (struct dos_header *) base;

  // kernel is put after the os loader by the boot loader
  doshdr = (struct dos_header *) (base + doshdr->e_cp * PAGESIZE);
  imghdr = (struct image_header *) ((char *) doshdr + doshdr->e_lfanew);

  // check that kernel image base matches the os base
  if (imghdr->optional.image_base != OSBASE) panic("kernel base address invalid");

  // calculate kernel size
  kernelsize = imghdr->optional.size_of_image;
  kernelpages = PAGES(kernelsize);
  kprintf("kernel size = %dK (%d pages)\n", kernelsize / K, kernelpages);

  // allocate page table for kernel
  if (kernelpages > PTES_PER_PAGE) panic("kernel too big");
  pt = (pte *) alloc_pages(1);
  pdir[L1IDX(OSBASE)] = (unsigned long) pt | PT_PRESENT | PT_WRITABLE;

  // allocate pages for kernel
  kerneladdr = alloc_pages(kernelpages);
  for (i = 0; i < kernelpages; i++) pt[i] = (unsigned long) (kerneladdr + i * PAGESIZE) | PT_PRESENT | PT_WRITABLE;

  // copy kernel to new address
  memcpy(kerneladdr, doshdr, kernelsize);

  // determine entry point for kernel
  krnlentry = imghdr->optional.address_of_entry_point + OSBASE;
}

void setup_descriptors()
{
  struct syspage *syspage;
  struct tss *tss;
  struct segment *seg;

  // get syspage virtual address
  syspage = (struct syspage *) SYSPAGE_ADDRESS;

  // initialize tss
  tss = &syspage->tss;
  tss->cr3 = (unsigned long) pdir;
  tss->eip = krnlentry;
  tss->cs = SEL_KTEXT;
  tss->ss0 =  SEL_KDATA;
  tss->ds =  SEL_KDATA;
  tss->es =  SEL_KDATA;
  tss->ss = SEL_KDATA;
  tss->esp = INITSTK_ADDRESS + PAGESIZE - 4;
  tss->esp0 = INITSTK_ADDRESS + PAGESIZE - 4;

  // setup kernel text segment (4GB read and execute, ring 0)
  seg = &syspage->gdt[GDT_KTEXT];
  seg->base_low = 0;
  seg->base_med = 0;
  seg->base_high = 0;
  seg->limit_low = 0xFFFF;
  seg->limit_high = 0xF;
  seg->access = D_CODE | D_DPL0 | D_READ | D_PRESENT;
  seg->granularity = D_BIG | D_BIG_LIM;

  // setup kernel data segment (4GB read and write, ring 0)
  seg = &syspage->gdt[GDT_KDATA];
  seg->base_low = 0;
  seg->base_med = 0;
  seg->base_high = 0;
  seg->limit_low = 0xFFFF;
  seg->limit_high = 0xF;
  seg->access = D_DATA | D_DPL0 | D_WRITE | D_PRESENT;
  seg->granularity = D_BIG | D_BIG_LIM;

  // setup user text segment (2GB read and execute, ring 3)
  seg = &syspage->gdt[GDT_UTEXT];
  seg->base_low = 0;
  seg->base_med = 0;
  seg->base_high = 0;
  seg->limit_low = ((OSBASE - 1) >> PAGESHIFT) & 0xFFFF;
  seg->limit_high = ((OSBASE - 1) >> (PAGESHIFT + 16)) & 0xF;
  seg->access = D_CODE | D_DPL3 | D_READ | D_PRESENT;
  seg->granularity = D_BIG | D_BIG_LIM;

  // setup user data segment (2GB read and write, ring 3)
  seg = &syspage->gdt[GDT_UDATA];
  seg->base_low = 0;
  seg->base_med = 0;
  seg->base_high = 0;
  seg->limit_low = ((OSBASE - 1) >> PAGESHIFT) & 0xFFFF;
  seg->limit_high = ((OSBASE - 1) >> (PAGESHIFT + 16)) & 0xF;
  seg->access = D_DATA | D_DPL3 | D_WRITE | D_PRESENT;
  seg->granularity = D_BIG | D_BIG_LIM;

  // setup tss segment
  seg = &syspage->gdt[GDT_TSS];
  seg->base_low = (unsigned short)((unsigned long) tss & 0xFFFF);
  seg->base_med = (unsigned char)(((unsigned long) tss >> 16) & 0xFF);
  seg->base_high = (unsigned char)(((unsigned long) tss >> 24) & 0xFF);
  seg->limit_low = sizeof(struct tss) - 1;
  seg->limit_high = 0;
  seg->access = D_TSS | D_DPL0 | D_PRESENT;
  seg->granularity = 0;

  // set gdt to the new segment descriptors
  gdtsel.limit = (sizeof(struct segment) * MAXGDT) - 1;
  gdtsel.dt = physgdt;
  __asm { lgdt gdtsel }

  // make all ldt selectors invalid
  ldtnull = 0;
  __asm { lldt [ldtnull] };

  // set idt to idt in syspage
  idtsel.limit = (sizeof(struct gate) * MAXIDT) - 1;
  idtsel.dt = physidt;
  __asm { lidt idtsel }

  // load task register with new TSS segment
  tssval = SEL_TSS;
  __asm { ltr tssval };
}

void __stdcall start(char *base, int reserved1, int reserved2)
{
  pte *pt;
  int i;

  // initialize video 
  init_video();
  clear_screen();
  kprintf("Starting OSLDR\n\n");

  // check that kernel does not overlap data section in osldr
  check_modules(base);

  // determine size of RAM
  mem_end = memsize();
  kprintf("Memory size is %d bytes, %d KB, %d MB\n", mem_end, mem_end / K, mem_end / M);
  kprintf("Base address is 0x%08X\n", base);

  // page alocation starts at 1MB
  heap = (char *) (1 * M);

  // allocate page for page directory
  pdir = (pte *) alloc_pages(1);

  // make recursive entry for access to pagetables
  pdir[L1IDX(PTBASE)] = (unsigned long) pdir | PT_PRESENT | PT_WRITABLE;

  // allocate system page
  syspage = (struct syspage *) alloc_pages(1);

  // allocate initial stack page
  initstk = alloc_pages(1);

  // allocate system page directory page
  syspagetable = (pte *) alloc_pages(1);

  // map system page, page directory and initial stack, and video buffer
  pdir[L1IDX(SYSBASE)] = (unsigned long) syspagetable | PT_PRESENT | PT_WRITABLE;
  syspagetable[L2IDX(SYSPAGE_ADDRESS)] = (unsigned long) syspage | PT_PRESENT | PT_WRITABLE;
  syspagetable[L2IDX(PAGEDIR_ADDRESS)] = (unsigned long) pdir | PT_PRESENT | PT_WRITABLE;
  syspagetable[L2IDX(INITSTK_ADDRESS)] = (unsigned long) initstk | PT_PRESENT | PT_WRITABLE;
  syspagetable[L2IDX(VIDBASE_ADDRESS)] = VIDEO_BASE | PT_PRESENT | PT_WRITABLE;

  // map first 4MB to physical memory
  pt = (pte *) alloc_pages(1);
  pdir[0] = (unsigned long) pt | PT_PRESENT | PT_WRITABLE;
  for (i = 0; i < PTES_PER_PAGE; i++) pt[i] = (i * PAGESIZE) | PT_PRESENT | PT_WRITABLE;

  // copy kernel to os base
  copy_kernel(base);

  // load tss, gtd and ldt
  //ldtnull = 0;
  //__asm { lldt [ldtnull] };

  // set page directory (CR3) enable paging (PG bit in CR0)
  __asm
  {
    mov eax, pdir
    mov cr3, eax
    mov eax, cr0
    or eax, 0x80000000
    mov cr0, eax
  }

  // setup new descriptors in syspage
  physgdt = &syspage->gdt;
  physidt = &syspage->idt;
  setup_descriptors();

  // reload segment registers
  __asm
  {
    mov	ax, SEL_KDATA
    mov	ds, ax
    mov	es, ax
    mov	fs, ax
    mov	gs, ax
    mov	ss, ax
    push SEL_KTEXT
    push offset xx1
    retf
  xx1:
  }

  kprintf("status: %dK heap used\n", ((unsigned long) heap - M) / K);

  // switch to inital kernel stack and jump to kernel
  __asm
  {
    mov esp, INITSTK_ADDRESS + PAGESIZE - 4
    push 0
    push 0
    push 0
    call [krnlentry]
    cli
    hlt
  }
}
